import http.client
import requests.structures
from ..exceptions import HttpUnknownHeaderObject, HttpUnknownResponseObject
from ..utils import get_url

"""
This module contains the base rule classes.
"""


class Rule:
    """
    Base class for rules. is_allowed and is_ready must be implemented.
    """

    def is_allowed(self, **kwargs):
        """Check if the request is allowed. Must be implemented.

        Args:
            **kwargs: Arbitrary keyword arguments to read args from.
        """
        raise NotImplementedError

    def is_ready(self):
        """Check if the rule is ready to be used."""
        raise NotImplementedError


class BulkRule(Rule):
    """
    Base class for bulk rules. filter_allowed and is_ready must be implemented.
    """

    def filter_allowed(self, **kwargs):
        """Filter a list of entries based on the rules in this evaluator."""
        raise NotImplementedError


class HttpRule(Rule):
    def __init__(self, user_agent=None):
        """Initialize the rule with a user agent.

        Args:
            user_agent (str): The user agent to pass on to the rules.
        """
        super().__init__()
        self.user_agent = user_agent

    def get_header_value(self, headers, header_name):
        """
        Handle the headers object to get the header value.

        Args:
            headers (dict|http.client.HTTPMessage|CaseInsensitiveDict): The headers object.
            header_name (str): The header name.

        Returns:
            str: The header value.
        """
        if type(headers) == dict or type(headers) == requests.structures.CaseInsensitiveDict:
            header_value = headers.get(header_name, "")
        elif type(headers) == list and len(headers) > 0 and type(headers[0]) == tuple:
            header_value = dict(headers).get(header_name, "")
        elif type(headers) == http.client.HTTPMessage:
            header_value = headers.get(header_name, "")
        else:
            raise HttpUnknownHeaderObject()

        return header_value

    def is_ready(self):
        """
        These rules should always be ready.
        """
        return True

    def _handle_url(self, url):
        """
        If given a raw URL, submit a request to get the image.

        Args:
            url (str): The URL of the resource.

        Returns:
            requests.Response: Response object.
        """
        return get_url(url, user_agent=self.user_agent)

    def get_header_value_from_response(self, response, header_name):
        """
        Handle the response object to get the header value.

        Args:
            response (http.client.HTTPResponse|requests.Response): The response object.
            header_name (str): The header name.

        Returns:
            str: The header value.
        """
        if type(response) == http.client.HTTPResponse:
            header_value = response.getheader(header_name, "")
        elif type(response) == requests.Response:
            header_value = response.headers.get(header_name, "")
        else:
            raise HttpUnknownResponseObject()
        return header_value



"""
Rules to manage validation using HTTP properties
"""

from ..exceptions import XRobotsTagNoParam, TDMRepNoParam
from .base import HttpRule


class XRobotsTagHeader(HttpRule):
    """
    This class wraps logic to read the X-Robots-Tag header.
    """
    AI_DISALLOWED_VALUES = ["noai", "noimageai"]
    INDEX_DISALLOWED_VALUES = ["noindex", "none", "noimageindex", "noai", "noimageai"]
    HEADER_NAME = "X-Robots-Tag"

    def __init__(self, user_agent=None, respect_noindex=False):
        """Create a new XRobotsTagHeader instance.

        Args:
            user_agent (str): The user agent to use when making requests to the Spawning AI API.
            respect_noindex (bool): If True, index rules will be respected alongside AI rules.
        """
        super().__init__(user_agent=user_agent)

        # index rules aren't for AI, so we ignore them by default.
        # They could have been delivered/found by any number of other means, even for internal use
        if respect_noindex:
            self.disallowed_headers = self.INDEX_DISALLOWED_VALUES
        else:
            self.disallowed_headers = self.AI_DISALLOWED_VALUES

    def is_allowed(self, url=None, response=None, headers=None, **kwargs):
        """Check if the X-Robots-Tag header allows the user agent to access the resource.

        Args:
            url: (str): The URL of the resource.
            response (http.client.HTTPResponse|requests.Response, optional): The response object. Defaults to None
            headers (dict|http.client.HTTPMessage, optional): The headers dictionary. Defaults to None.

        Returns:
            bool: True if the user agent is allowed to access the resource, False otherwise.
        """

        if headers:
            header_value = self.get_header_value(headers, self.HEADER_NAME)
        elif response:
            header_value = self.get_header_value_from_response(response, self.HEADER_NAME)
        elif url:
            response = self._handle_url(url)
            header_value = self.get_header_value(response.headers, self.HEADER_NAME)
        else:
            raise XRobotsTagNoParam()

        return self._eval_header_value(header_value, **kwargs)

    def _eval_header_value(self, header_value, user_agent=None, **kwargs):
        """
        Evaluate the header value to determine if the user agent is allowed to access the resource.

        Args:
            header_value (str): The header value.
            user_agent (str): Override user agent to use when making requests to the Spawning AI API.

        Returns:
            bool: True if the user agent is allowed to access the resource, False otherwise.
        """
        if not header_value:
            return True

        # if we have a specific user agent
        if not user_agent:
            user_agent = self.user_agent

        # check if blocking all user agents
        for value in header_value.split(","):
            if value.strip() in self.disallowed_headers:
                return False

            # check if blocking specific user agent
            if user_agent:
                ua_values = value.split(":")
                if len(ua_values) == 2 and ua_values[0].strip() == user_agent \
                        and ua_values[1].strip() in self.disallowed_headers:
                    return False

        return True


class TDMRepHeader(HttpRule):
    """
    This class wraps logic to evaluate the TDM Reservation Protocol headers: https://www.w3.org/2022/tdmrep/.
    """
    HEADER_NAME = "tdm-reservation"

    def __init__(self):
        """Create a new TDMRepHeaders instance."""
        super().__init__()

    def is_allowed(self, url=None, response=None, headers=None, **kwargs):
        """Check if the tdm-rep header allows access to the resource without a policy.

        Args:
            url: (str): The URL of the resource.
            response (http.client.HTTPResponse|requests.Response, optional): The response object. Defaults to None
            headers (dict|http.client.HTTPMessage, optional): The headers dictionary. Defaults to None.

        Returns:
            bool: True if access is allowed for the resource, False otherwise.
        """

        if headers:
            header_value = self.get_header_value(headers, self.HEADER_NAME)
        elif response:
            header_value = self.get_header_value_from_response(response, self.HEADER_NAME)
        elif url:
            response = self._handle_url(url)
            header_value = self.get_header_value(response.headers, self.HEADER_NAME)
        else:
            raise TDMRepNoParam()

        return self._eval_header_value(header_value, **kwargs)

    def _eval_header_value(self, header_value, **kwargs):
        """
        Evaluate the header value to determine if the resource permits anonymous access.

        Args:
            header_value (str): The header value.

        Returns:
            bool: True if resource allows access without a policy, False otherwise.
        """

        if not header_value:
            return True

        print("HERE")
        print(header_value)
        return header_value.strip() != "1"




import requests
import urllib.request
from unittest import TestCase
import datadiligence as dd
from datadiligence.rules import TDMRepHeader
import time

# starting local server to echo back headers
from werkzeug.serving import make_server
from server.app import app
import threading


class TDMRepTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = make_server('localhost', 5001, app)
        cls.server_thread = threading.Thread(target=cls.server.serve_forever)
        cls.server_thread.start()
        time.sleep(1)  # wait for server to start

        cls.rule = TDMRepHeader()

    def test_noheader(self):
        self.assertTrue(self.rule._eval_header_value(""))
        self.assertTrue(self.rule._eval_header_value(None))

    def test_tdm_block(self):
        self.assertFalse(self.rule._eval_header_value("1"))
        self.assertTrue(self.rule._eval_header_value("0"))
        self.assertTrue(self.rule._eval_header_value("other"))

    def test_stdlib(self):
        request = urllib.request.Request("http://localhost:5001/tdmrep", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertEqual(self.rule.get_header_value_from_response(response, self.rule.HEADER_NAME), "0")
            self.assertEqual(self.rule.get_header_value(response.headers, self.rule.HEADER_NAME), "0")
            self.assertEqual(self.rule.get_header_value(response.getheaders(), self.rule.HEADER_NAME), "0")
            self.assertTrue(self.rule.is_allowed(response=response))
            self.assertTrue(self.rule.is_allowed(headers=response.headers))

        request = urllib.request.Request("http://localhost:5001/blocktdmrep", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertEqual(self.rule.get_header_value_from_response(response, self.rule.HEADER_NAME), "1")
            self.assertEqual(self.rule.get_header_value(response.headers, self.rule.HEADER_NAME), "1")
            self.assertFalse(self.rule.is_allowed(response=response))
            self.assertFalse(self.rule.is_allowed(headers=response.headers))

    def test_requests_lib(self):
        response = requests.get("http://localhost:5001/tdmrep", timeout=3)
        self.assertEqual(self.rule.get_header_value_from_response(response, self.rule.HEADER_NAME), "0")
        self.assertEqual(self.rule.get_header_value(response.headers, self.rule.HEADER_NAME), "0")
        self.assertTrue(self.rule.is_allowed(response=response))
        self.assertTrue(self.rule.is_allowed(headers=response.headers))

        response = requests.get("http://localhost:5001/blocktdmrep")
        self.assertEqual(self.rule.get_header_value_from_response(response, self.rule.HEADER_NAME), "1")
        self.assertEqual(self.rule.get_header_value(response.headers, self.rule.HEADER_NAME), "1")
        self.assertFalse(self.rule.is_allowed(response=response))
        self.assertFalse(self.rule.is_allowed(headers=response.headers))

    def test_exceptions(self):
        self.assertRaises(dd.exceptions.TDMRepNoParam, self.rule.is_allowed, None, None)
        self.assertRaises(dd.exceptions.HttpUnknownHeaderObject, self.rule.get_header_value, None, None)
        self.assertRaises(dd.exceptions.HttpUnknownResponseObject, self.rule.get_header_value_from_response, None, None)

    def test_url_arg(self):
        self.assertTrue(self.rule.is_allowed(url="http://localhost:5001/tdmrep"))
        self.assertFalse(self.rule.is_allowed(url="http://localhost:5001/blocktdmrep"))

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server_thread.join()
